"""Provides an abstract base class Aggregator and multiple sub-classes"""

from abc import ABC, abstractmethod
from typing import List

import numpy as np
from numpy import ndarray

from xain_fl.logger import StructLogger, get_logger

logger: StructLogger = get_logger(__name__)


class Aggregator(ABC):  # pylint: disable=too-few-public-methods
    """Abstract base class for model weights aggregation strategies."""

    def __init__(self):
        """Initialize the aggregator."""

    @abstractmethod
    def aggregate(
        self, multiple_model_weights: List[ndarray], aggregation_data: List[int]
    ) -> ndarray:
        """Aggregate the weights of multiple models.

        Args:
            multiple_model_weights (List[ndarray]): The weights of multiple models.
            aggregation_data (List[int]): Meta data for model aggregation.

        Returns:
            ndarray: The aggregated model weights.
        """

        raise NotImplementedError()


class IdentityAggregator(Aggregator):  # pylint: disable=too-few-public-methods
    """Identity aggregation."""

    def aggregate(
        self, multiple_model_weights: List[ndarray], aggregation_data: List[int]
    ) -> ndarray:
        """Identity aggregation only for one set of model weights.

        Args:
            multiple_model_weights (List[ndarray]): The weights of multiple models. Must have only
                one set of weights.
            aggregation_data (List[int]): Meta data for model aggregation. Not used here.

        Returns:
            ndarray: The identical model weights.

        Raises:
            ValueError: If more than one set of model weights is provided.
        """

        if len(multiple_model_weights) > 1:
            raise ValueError("Invalid number of model weights!")

        return multiple_model_weights[0]


class ModelSumAggregator(Aggregator):  # pylint: disable=too-few-public-methods
    """Summation of models aggregation."""

    def aggregate(
        self, multiple_model_weights: List[ndarray], aggregation_data: List[int]
    ) -> ndarray:
        """Aggregate the weights of multiple models by summation.

        Args:
            multiple_model_weights (List[ndarray]): The weights of multiple models.
            aggregation_data (List[int]): Meta data for model aggregation. Not used here.

        Returns:
            ndarray: The aggregated model weights.
        """

        return sum(multiple_model_weights)


class WeightedAverageAggregator(Aggregator):  # pylint: disable=too-few-public-methods
    """Weighted average aggregation."""

    def aggregate(
        self, multiple_model_weights: List[ndarray], aggregation_data: List[int]
    ) -> ndarray:
        """Aggregate the weights of multiple models by weighted averages.

        Proposed by McMahan et al in: https://arxiv.org/abs/1602.05629

        Args:
            multiple_model_weights (List[ndarray]): The weights of multiple models.
            aggregation_data (List[int]): Meta data for model aggregation. Here it is expected to be
                the number of train samples per set of model weights. If all numbers are zero, e.g.
                in a 0th round for weight initialization, then all model weights are weighted
                equally.

        Returns:
            ndarray: The aggregated model weights.
        """

        # get aggregation weights
        aggregation_weights: ndarray
        if any(aggregation_data):
            aggregation_weights = np.array(aggregation_data) / np.sum(aggregation_data)
        else:
            aggregation_weights = np.ones_like(aggregation_data) / len(aggregation_data)

        # compute weighted average
        scaled_model_weights: List[ndarray] = [
            model_weight * aggregation_weight
            for model_weight, aggregation_weight in zip(
                multiple_model_weights, aggregation_weights
            )
        ]
        aggregated_model_weights: ndarray = np.sum(scaled_model_weights, axis=0)

        return aggregated_model_weights


# TODO: (XP-351) decide how to continue with that
# def federated_averaging(multiple_model_weights: List[ndarray], weighting: ndarray) -> ndarray:
#     """Calculates weighted averages of provided list of weight updates, as proposed by McMahan et
#         al. in: https://arxiv.org/abs/1602.05629
#
#     Args:
#         multiple_model_weights (List[ndarray]): List of model weight.
#         weighting (ndarray): Describes relative weight of each model weight. Required to be the
#             same length as argument multiple_model_weights.
#
#     Returns:
#         ndarray: The aggregated model weights.
#     """
#
#     assert weighting.ndim == 1
#     assert len(multiple_model_weights) == weighting.shape[0]
#
#     model_weights_avg: ndarray = multiple_model_weights[0] * weighting[0]
#
#     # Aggregate (weighted) updates
#     for weights, update_weighting in zip(multiple_model_weights[1:], weighting[1:]):
#         model_weights_avg += update_weighting * weights
#
#     weighting_sum = np.sum(weighting)
#     model_weights_avg /= weighting_sum
#
#     return model_weights_avg

# class EvoAgg(Aggregator):
#     """Experimental"""
#
#     def __init__(self, evaluator: Evaluator):
#         super().__init__()
#         self.evaluator = evaluator
#
#     def aggregate(self, weight_updates: List[Tuple[List[ndarray], int]]) -> List[ndarray]:
#         weight_matrices = [model_weights for model_weights, num_examples in weight_updates]
#         return evo_agg(weight_matrices, self.evaluator)

# def evo_agg(weight_updates: List[ndarray], evaluator: Evaluator) -> ndarray:
#     """Experimental
#
#         - Init different weightings
#         - Aggregate weight updates according to those weightings ("candidates")
#         - Evaluate all candidates on the validation set
#         - Pick (a) best candidate, or (b) average of n best candidates
#     """
#     # Compute candidates
#     # TODO in parallel, do:
#     weights_prime_candidates = []
#     for _ in range(3):
#         candidate = _compute_candidate(weight_updates, evaluator)
#         logger.debug("Candidate metadata", weighting=candidate[0], loss=candidate[2])
#
#         weights_prime_candidates.append(candidate)
#     # Return best candidate
#     best_candidate = _pick_best_candidate(weights_prime_candidates)
#     return best_candidate


# def _pick_best_candidate(candidates: List) -> List[ndarray]:
#     _, best_candidate, best_loss, _ = candidates[0]
#     for _, candidate, loss, _ in candidates[1:]:
#         if loss < best_loss:
#             best_candidate = candidate
#             best_loss = loss
#     return best_candidate


# def _compute_candidate(
#     weight_updates: List[ndarray], evaluator: Evaluator
# ) -> Tuple[ndarray, ndarray, float, float]:
#     weighting = _random_weighting(len(weight_updates))
#     weights_prime_candidate = federated_averaging(weight_updates, weighting)
#     loss, acc = evaluator.evaluate(weights_prime_candidate)
#     return weighting, weights_prime_candidate, loss, acc


# def _random_weighting(num_weightings: int, low=0.5, high=1.5) -> ndarray:
#     return np.random.uniform(low=low, high=high, size=(num_weightings,))



"""Defines constants to be used throughout the module
"""

# TODO: move to `xain_fl.config.py`
# once https://xainag.atlassian.net/browse/XP-265 is done
_ONE_DAY_IN_SECONDS: int = 60 * 60 * 24
HEARTBEAT_TIME: int = 10
HEARTBEAT_TIMEOUT: int = 5



"""XAIN FL conftest for cproto"""

from concurrent import futures
import threading

import grpc
import pytest
from xain_proto.fl import coordinator_pb2_grpc

from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.coordinator.coordinator_grpc import CoordinatorGrpc
from xain_fl.coordinator.heartbeat import monitor_heartbeats
from xain_fl.fl.coordinator.aggregate import ModelSumAggregator
from xain_fl.fl.coordinator.controller import IdController

from .port_forwarding import ConnectionManager


@pytest.fixture
def coordinator_service():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    coordinator = Coordinator(
        minimum_participants_in_round=10, fraction_of_participants=1.0
    )
    coordinator_grpc = CoordinatorGrpc(coordinator)
    coordinator_pb2_grpc.add_CoordinatorServicer_to_server(coordinator_grpc, server)
    server.add_insecure_port("localhost:50051")
    server.start()
    yield coordinator_grpc
    server.stop(0)


@pytest.fixture
def mock_coordinator_service():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    agg = ModelSumAggregator()
    ctrl = IdController()
    coordinator = Coordinator(
        num_rounds=2,
        minimum_participants_in_round=1,
        fraction_of_participants=1.0,
        aggregator=agg,
        controller=ctrl,
    )
    coordinator_grpc = CoordinatorGrpc(coordinator)
    coordinator_pb2_grpc.add_CoordinatorServicer_to_server(coordinator_grpc, server)
    server.add_insecure_port("localhost:50051")
    server.start()
    terminate_event = threading.Event()
    monitor_thread = threading.Thread(
        target=monitor_heartbeats, args=(coordinator, terminate_event)
    )
    monitor_thread.start()
    yield coordinator_grpc
    terminate_event.set()
    monitor_thread.join()
    server.stop(0)


@pytest.fixture
def participant_stub():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Returns:
        [type]: [description]
    """

    channel = grpc.insecure_channel("localhost:50051")
    stub = coordinator_pb2_grpc.CoordinatorStub(channel)

    yield stub

    channel.close()


def port_generator():
    """A generator that yields incrementing port numbers.

    """
    port = 50052
    while True:
        yield port
        port += 1


@pytest.fixture
def participant_stubs():
    """Generator that yields functions instantiate participant stubs.
    Each participant creates a new TCP connection, so that they get a
    different participant ID.

    """

    ports = port_generator()
    connections = ConnectionManager()
    channels = []

    def generate_participant_stubs():
        for port in ports:
            connections.start("localhost", port, "localhost", 50051)
            channel = grpc.insecure_channel(f"localhost:{port}")
            channels.append(channel)
            stub = coordinator_pb2_grpc.CoordinatorStub(channel)
            yield stub

    yield generate_participant_stubs()

    for channel in channels:
        channel.close()

    connections.stop_all()



"""XAIN FL serving"""

from concurrent import futures
import threading

import grpc
from xain_proto.fl import coordinator_pb2_grpc

from xain_fl.config import ServerConfig
from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.coordinator.coordinator_grpc import CoordinatorGrpc
from xain_fl.coordinator.heartbeat import monitor_heartbeats
from xain_fl.logger import StructLogger, get_logger

logger: StructLogger = get_logger(__name__)


def serve(coordinator: Coordinator, server_config: ServerConfig) -> None:
    """Start a coordinator service and keep it running until an
    interruption signal (``SIGINT``) is received.

    Args:

        coordinator:
            :class:`xain_fl.coordinator.coordinator.Coordinator`
            instance to run

        store:
            client for the storage service used by the coordinator to
            upload the aggregated weights and retrieve the
            participants weights.

        server_config:
            server configuration: binding address, gRPC options, etc.

    """
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10), options=server_config.grpc_options
    )
    coordinator_pb2_grpc.add_CoordinatorServicer_to_server(
        CoordinatorGrpc(coordinator), server
    )
    server.add_insecure_port(f"{server_config.host}:{server_config.port}")
    server.start()

    logger.info("Coordinator waiting for connections...")

    terminate_event = threading.Event()
    try:
        monitor_heartbeats(coordinator, terminate_event)
    except KeyboardInterrupt:
        terminate_event.set()
        server.stop(0)



"""XAIN FL tests for controller"""

import numpy as np

from xain_fl.fl.coordinator.controller import RandomController


def test_random_controller():
    """Tests that the length of selected ids is correct and that
    there's no replacement.
    """
    participant_ids = ["a", "b", "c", "d", "e", "f", "g"]
    fractions = np.arange(0.25, 1, 0.25)
    expected_lengths = [
        np.ceil(fraction * len(participant_ids)) for fraction in fractions
    ]
    for fraction, expected_length in zip(fractions, expected_lengths):
        controller = RandomController(fraction_of_participants=fraction)
        ids = controller.select_ids(participant_ids)
        set_ids = set(ids)

        # check that length of set_ids is as expected
        assert len(set_ids) == expected_length

        # check that every element of set_ids belongs to participant_ids
        assert set_ids.issubset(participant_ids)



"""XAIN FL Participants"""

import threading
import time
from typing import Dict, List

from xain_fl.coordinator import HEARTBEAT_TIME, HEARTBEAT_TIMEOUT


class ParticipantContext:  # pylint: disable=too-few-public-methods
    """Class to store state about each participant. Currently it only stores the `participant_id`
    and the time when the next heartbeat_expires.

    In the future we may store more information like in what state a participant is in e.g.
    `IDLE`, `RUNNING`, ...

    Args:

        participant_id: The id of the participant. Typically a
            host:port or public key when using SSL.
    """

    def __init__(self, participant_id: str) -> None:
        self.participant_id: str = participant_id
        self.heartbeat_expires: float = time.time() + HEARTBEAT_TIME + HEARTBEAT_TIMEOUT


class Participants:
    """This class provides some useful methods to handle all the participants
    connected to a coordinator in a thread safe manner by protecting access to
    the participants list with a lock.
    """

    def __init__(self) -> None:
        self.participants: Dict[str, ParticipantContext] = {}
        self._lock: threading.Lock = threading.Lock()

    def add(self, participant_id: str) -> None:
        """Adds a new participant to the list of participants.

        Args:

            participant_id: The id of the participant to add.
        """

        with self._lock:
            self.participants[participant_id] = ParticipantContext(participant_id)

    def remove(self, participant_id: str) -> None:
        """Removes a participant from the list of participants.

        This will be typically used after a participant is
        disconnected from the coordinator.

        Args:

            participant_id: The id of the participant to remove.
        """

        with self._lock:
            if participant_id in self.participants:
                del self.participants[participant_id]

    def next_expiration(self) -> float:
        """Helper method to check what is the next heartbeat to expire.

        Currently being used by the `heartbeat_monitor` to check how long it should sleep until
        the next check.

        Returns:

            The next heartbeat to expire.
        """

        with self._lock:
            if self.participants:
                return min([p.heartbeat_expires for p in self.participants.values()])

        return time.time() + HEARTBEAT_TIME + HEARTBEAT_TIMEOUT

    def len(self) -> int:
        """Get the number of participants.

        Returns:

            The number of participants in the list.
        """

        with self._lock:
            return len(self.participants)

    def ids(self) -> List[str]:
        """Get the ids of the participants.

        Returns:

            The list of participant ids.
        """

        with self._lock:
            return list(self.participants.keys())

    def update_expires(self, participant_id: str) -> None:
        """Updates the heartbeat expiration time for a participant.

        This is currently called by the :class:`xain_fl.coordinator.coordinator.Coordinator`
        every time a participant sends a heartbeat.

        Args:

            participant_id: The id of the participant to update the expire time.
        """

        with self._lock:
            self.participants[participant_id].heartbeat_expires = (
                time.time() + HEARTBEAT_TIME + HEARTBEAT_TIMEOUT
            )



"""XAIN FL structured logger"""

import logging
from typing import Any, Optional, Union

import structlog

StructLogger = (
    structlog._config.BoundLoggerLazyProxy  # pylint: disable=protected-access
)


def configure_aimetrics_logger():
    """Configure a logger named "aimetrics" with a configurable log
    level.

    """
    AIMETRICS = 25  # pylint: disable=invalid-name
    structlog.stdlib.AIMETRICS = AIMETRICS
    structlog.stdlib._NAME_TO_LEVEL[  # pylint: disable=protected-access
        "aimetrics"
    ] = AIMETRICS
    structlog.stdlib._LEVEL_TO_NAME[  # pylint: disable=protected-access
        AIMETRICS
    ] = "aimetrics"
    logging.addLevelName(AIMETRICS, "aimetrics")

    def aimetrics(self, msg: str, *args: Any, **kw: Any) -> Any:
        return self.log(AIMETRICS, msg, *args, **kw)

    structlog.stdlib._FixedFindCallerLogger.aimetrics = (  # pylint: disable=protected-access
        aimetrics
    )
    structlog.stdlib.BoundLogger.aimetrics = aimetrics


def configure_structlog():
    """Set the structlog configuration"""
    structlog.configure(
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def set_log_level(level: Union[str, int]):
    """Set the log level on the root logger. Since by default, the root
    logger log level is inherited by all the loggers, this is like
    setting a default log level.

    Args:

        level: the log level, as documented in the `Python standard
            library <https://docs.python.org/3/library/logging.html#levels>`_
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(level)


def initialize_logging():
    """Set up logging

    """
    configure_aimetrics_logger()
    configure_structlog()


def get_logger(
    name: str, level: Optional[int] = None
) -> structlog._config.BoundLoggerLazyProxy:  # pylint: disable=protected-access
    """Wrap python logger with default configuration of structlog.

    Args:
        name: Identification name. For module name pass ``name=__name__``.
        level: Threshold for this logger.

    Returns:
        Wrapped python logger with default configuration of structlog.
    """
    formatter = structlog.stdlib.ProcessorFormatter(
        processor=structlog.processors.JSONRenderer(indent=2, sort_keys=True)
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.addHandler(handler)

    if level is not None:
        logger.setLevel(level)

    return structlog.wrap_logger(logger)



"""XAIN FL gRPC Coordinator"""

import grpc
from xain_proto.fl.coordinator_pb2 import (
    EndTrainingRoundRequest,
    EndTrainingRoundResponse,
    HeartbeatRequest,
    HeartbeatResponse,
    RendezvousRequest,
    RendezvousResponse,
    StartTrainingRoundRequest,
    StartTrainingRoundResponse,
)
from xain_proto.fl.coordinator_pb2_grpc import CoordinatorServicer

from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.tools.exceptions import (
    DuplicatedUpdateError,
    InvalidRequestError,
    UnknownParticipantError,
)


class CoordinatorGrpc(CoordinatorServicer):
    """The Coordinator gRPC service.

    The main logic for the Coordinator is decoupled from gRPC and implemented in the
    :class:`xain_fl.coordinator.coordinator.Coordinator` class. The gRPC message only handles
    client requests and forwards the messages to
    :class:`xain_fl.coordinator.coordinator.Coordinator`.

    Args:

        coordinator: The Coordinator state machine.
    """

    def __init__(self, coordinator: Coordinator):
        self.coordinator: Coordinator = coordinator

    def Rendezvous(
        self, request: RendezvousRequest, context: grpc.ServicerContext
    ) -> RendezvousResponse:
        """The Rendezvous gRPC method.

        A participant contacts the coordinator and the coordinator adds the
        participant to its list of participants. If the coordinator already has
        all the participants it tells the participant to try again later.

        Args:

            request: The participant's request.
            context: The context associated with the gRPC request.

        Returns:

            The response to the participant's request. The response is
            an enum containing either:

                - `ACCEPT`: If the
                  :class:`xain_fl.coordinator.coordinator.Coordinator`
                  does not have enough participants.

                - `LATER`: If the
                    :class:`xain_fl.coordinator.coordinator.Coordinator`
                    already has enough participants.
        """
        return self.coordinator.on_message(request, context.peer())

    def Heartbeat(
        self, request: HeartbeatRequest, context: grpc.ServicerContext
    ) -> HeartbeatResponse:
        """The Heartbeat gRPC method.

        Participants periodically send an heartbeat so that the
        :class:`Coordinator` can detect failures.

        Args:

            request: The participant's request. The participant's
                request contains the current
                :class:`~.coordinator_pb2.State` and round number the
                participant is on.

            context: The context associated with the gRPC request.

        Returns:

            The response to the participant's request. The response
            contains both the :class:`~.coordinator_pb2.State` and the
            current round the coordinator is on. If a training session
            has not started yet the round number defaults to 0.
        """
        try:
            return self.coordinator.on_message(request, context.peer())
        except UnknownParticipantError as error:
            context.set_details(str(error))
            context.set_code(grpc.StatusCode.PERMISSION_DENIED)
            return HeartbeatResponse()

    def StartTrainingRound(
        self, request: StartTrainingRoundRequest, context: grpc.ServicerContext,
    ) -> StartTrainingRoundResponse:
        """The StartTrainingRound gRPC method.

        Once a participant is notified that the :class:`xain_fl.coordinator.coordinator.Coordinator`
        is in a round (through the state advertised in the
        :class:`~.coordinator_pb2.HeartbeatResponse`), the participant should call this
        method in order to get the global model weights in order to start the
        training for the round.

        Args:

            request: The participant's request.
            context: The context associated with the gRPC request.

        Returns:

            The response to the participant's request. The response
            contains the global model weights.
        """
        try:
            return self.coordinator.on_message(request, context.peer())
        except UnknownParticipantError as error:
            context.set_details(str(error))
            context.set_code(grpc.StatusCode.PERMISSION_DENIED)
            return StartTrainingRoundResponse()
        except InvalidRequestError as error:
            context.set_details(str(error))
            context.set_Code(grpc.StatusCode.FAILED_PRECONDITION)
            return StartTrainingRoundResponse()

    def EndTrainingRound(
        self, request: EndTrainingRoundRequest, context: grpc.ServicerContext
    ) -> EndTrainingRoundResponse:
        """The EndTrainingRound gRPC method.

        Once a participant has finished the training for the round it calls this
        method in order to submit to the :class:`xain_fl.coordinator.coordinator.Coordinator`
        the updated weights.

        Args:

            request: The participant's request. The request contains
                the updated weights as a result of the training as
                well as any metrics helpful for the
                :class:`xain_fl.coordinator.coordinator.Coordinator`.

            context: The context associated with the gRPC request.

        Returns:

            The response to the participant's request. The response is
            just an acknowledgment that the
            :class:`xain_fl.coordinator.coordinator.Coordinator`
            successfully received the updated weights.
        """
        try:
            return self.coordinator.on_message(request, context.peer())
        except DuplicatedUpdateError as error:
            context.set_details(str(error))
            context.set_code(grpc.StatusCode.ALREADY_EXISTS)
            return EndTrainingRoundResponse()
        except UnknownParticipantError as error:
            context.set_details(str(error))
            context.set_code(grpc.StatusCode.PERMISSION_DENIED)
            return EndTrainingRoundResponse()



"""XAIN FL tests for coordinator"""

import numpy as np
import pytest
from xain_proto.fl.coordinator_pb2 import (
    EndTrainingRoundRequest,
    HeartbeatRequest,
    RendezvousReply,
    RendezvousRequest,
    RendezvousResponse,
    StartTrainingRoundRequest,
    State,
)
from xain_proto.np import proto_to_ndarray

from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.tools.exceptions import (
    DuplicatedUpdateError,
    InvalidRequestError,
    UnknownParticipantError,
)


def test_rendezvous_accept():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    coordinator: Coordinator = Coordinator()
    response: RendezvousResponse = coordinator.on_message(
        RendezvousRequest(), "participant1"
    )

    assert isinstance(response, RendezvousResponse)
    assert response.reply == RendezvousReply.ACCEPT


def test_rendezvous_later_fraction_1():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=1.0
    )
    coordinator.on_message(RendezvousRequest(), "participant1")
    response = coordinator.on_message(RendezvousRequest(), "participant2")

    assert isinstance(response, RendezvousResponse)
    assert response.reply == RendezvousReply.LATER


def test_rendezvous_later_fraction_05():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=0.5
    )

    # with 0.5 fraction it needs to accept at least two participants
    response = coordinator.on_message(RendezvousRequest(), "participant1")

    assert isinstance(response, RendezvousResponse)
    assert response.reply == RendezvousReply.ACCEPT

    response = coordinator.on_message(RendezvousRequest(), "participant2")

    assert isinstance(response, RendezvousResponse)
    assert response.reply == RendezvousReply.ACCEPT

    # the third participant must receive LATER RendezvousReply
    response = coordinator.on_message(RendezvousRequest(), "participant3")

    assert isinstance(response, RendezvousResponse)
    assert response.reply == RendezvousReply.LATER


def test_coordinator_state_standby_round():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # tests that the coordinator transitions from STANDBY to ROUND once enough participants
    # are connected
    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=1.0
    )

    assert coordinator.state == State.STANDBY

    coordinator.on_message(RendezvousRequest(), "participant1")

    assert coordinator.state == State.ROUND
    assert coordinator.current_round == 0


def test_start_training_round():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    test_weights = np.arange(10)
    coordinator = Coordinator(
        minimum_participants_in_round=1,
        fraction_of_participants=1.0,
        weights=test_weights,
    )
    coordinator.on_message(RendezvousRequest(), "participant1")

    response = coordinator.on_message(StartTrainingRoundRequest(), "participant1")
    received_weights = proto_to_ndarray(response.weights)

    np.testing.assert_equal(test_weights, received_weights)


def start_training_round_wrong_state():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # if the coordinator receives a StartTrainingRound request while not in the
    # ROUND state it will raise an exception
    coordinator = Coordinator(
        minimum_participants_in_round=2, fraction_of_participants=1.0
    )
    coordinator.on_message(RendezvousRequest(), "participant1")

    with pytest.raises(InvalidRequestError):
        coordinator.on_message(StartTrainingRoundRequest(), "participant1")


def test_end_training_round():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # we need two participants so that we can check the status of the local update mid round
    # with only one participant it wouldn't work because the local updates state is cleaned at
    # the end of each round
    coordinator = Coordinator(
        minimum_participants_in_round=2, fraction_of_participants=1.0
    )
    coordinator.on_message(RendezvousRequest(), "participant1")
    coordinator.on_message(RendezvousRequest(), "participant2")

    coordinator.on_message(EndTrainingRoundRequest(), "participant1")

    assert len(coordinator.round.updates) == 1


def test_end_training_round_update():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # Test that the round number is updated once all participants sent their updates
    coordinator = Coordinator(
        minimum_participants_in_round=2, fraction_of_participants=1.0, num_rounds=2
    )
    coordinator.on_message(RendezvousRequest(), "participant1")
    coordinator.on_message(RendezvousRequest(), "participant2")

    # check that we are currently in round 0
    assert coordinator.current_round == 0

    coordinator.on_message(EndTrainingRoundRequest(), "participant1")
    # check we are still in round 0
    assert coordinator.current_round == 0
    coordinator.on_message(EndTrainingRoundRequest(), "participant2")

    # check that round number was updated
    assert coordinator.current_round == 1


def test_end_training_round_reinitialize_local_models():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    coordinator = Coordinator(
        minimum_participants_in_round=2, fraction_of_participants=1.0, num_rounds=2
    )
    coordinator.on_message(RendezvousRequest(), "participant1")
    coordinator.on_message(RendezvousRequest(), "participant2")

    coordinator.on_message(EndTrainingRoundRequest(), "participant1")

    # After one participant sends its updates we should have one update in the coordinator
    assert len(coordinator.round.updates) == 1

    coordinator.on_message(EndTrainingRoundRequest(), "participant2")

    # once the second participant delivers its updates the round ends and the local models
    # are reinitialized
    assert coordinator.round.updates == {}


def test_training_finished():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=1.0, num_rounds=2
    )
    coordinator.on_message(RendezvousRequest(), "participant1")

    # Deliver results for 2 rounds
    coordinator.on_message(EndTrainingRoundRequest(), "participant1")
    coordinator.on_message(EndTrainingRoundRequest(), "participant1")

    assert coordinator.state == State.FINISHED


def test_wrong_participant():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # coordinator should not accept requests from participants that it has not accepted
    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=1.0
    )
    coordinator.on_message(RendezvousRequest(), "participant1")

    with pytest.raises(UnknownParticipantError):
        coordinator.on_message(HeartbeatRequest(), "participant2")

    with pytest.raises(UnknownParticipantError):
        coordinator.on_message(StartTrainingRoundRequest(), "participant2")

    with pytest.raises(UnknownParticipantError):
        coordinator.on_message(EndTrainingRoundRequest(), "participant2")


def test_duplicated_update_submit():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # the coordinator should not accept multiples updates from the same participant
    # in the same round
    coordinator = Coordinator(
        minimum_participants_in_round=2, fraction_of_participants=1.0
    )
    coordinator.on_message(RendezvousRequest(), "participant1")
    coordinator.on_message(RendezvousRequest(), "participant2")

    coordinator.on_message(EndTrainingRoundRequest(), "participant1")

    with pytest.raises(DuplicatedUpdateError):
        coordinator.on_message(EndTrainingRoundRequest(), "participant1")


def test_remove_participant():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=1.0
    )
    coordinator.on_message(RendezvousRequest(), "participant1")

    assert coordinator.state == State.ROUND

    coordinator.remove_participant("participant1")

    assert coordinator.participants.len() == 0
    assert coordinator.state == State.STANDBY

    coordinator.on_message(RendezvousRequest(), "participant1")

    assert coordinator.state == State.ROUND


def test_number_of_selected_participants():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # test that the coordinator needs minimum 3 participants and selects 2 of them
    coordinator = Coordinator(
        minimum_participants_in_round=2, fraction_of_participants=0.6
    )
    coordinator.on_message(RendezvousRequest(), "participant1")

    # the coordinator should wait for three participants to be connected before starting a round,
    # and select participants. Before that coordinator.round.participant_ids is an empty list
    assert coordinator.minimum_connected_participants == 3
    assert coordinator.state == State.STANDBY
    assert coordinator.round.participant_ids == []

    coordinator.on_message(RendezvousRequest(), "participant2")

    assert coordinator.state == State.STANDBY
    assert coordinator.round.participant_ids == []

    # add the third participant
    coordinator.on_message(RendezvousRequest(), "participant3")

    # now the coordinator must have started a round and selected 2 participants
    assert coordinator.state == State.ROUND
    assert len(coordinator.round.participant_ids) == 2


def test_correct_round_advertised_to_participants():
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    # test that only selected participants receive ROUND state and the others STANDBY
    coordinator = Coordinator(
        minimum_participants_in_round=1, fraction_of_participants=0.5
    )
    coordinator.on_message(RendezvousRequest(), "participant1")
    coordinator.on_message(RendezvousRequest(), "participant2")

    # override selected participant
    coordinator.round.participant_ids = ["participant1"]

    # state ROUND will be advertised to participant1 (which has been selected)
    response = coordinator.on_message(HeartbeatRequest(), "participant1")
    assert response.state == State.ROUND

    # state STANDBY will be advertised to participant2 (which has NOT been selected)
    response = coordinator.on_message(HeartbeatRequest(), "participant2")
    assert response.state == State.STANDBY



"""This module provides helpers for reading and validating the TOML
configuration.

"""

# Implementation notes:
# =====================
#
# We use the `schema` library to validate the configuration provided
# by the user. However, `schema` works mainly with dictionaries, which
# are not particularly convenient for us:
#
# - we cannot generate documentation for all the dictionary keys
# - the syntax for accessing values is quite verbose, especially if
#   the dictionary has nested values
#
# Thus, although we use schemas to validate configurations (AI_SCHEMA,
# SERVER_SCHEMA, and STORAGE_SCHEMA), we don't expose them. Instead,
# we use them to dynamically generate classes where attributes are the
# schema keys: AiConfig, ServerConfig, and StorageConfig. This
# hackery happens in create_class_from_schema(). It is not
# perfect. For instance, we cannot document the type of each
# attribute. But it is arguably better than using plain dictionaries.

from collections import namedtuple
import ipaddress
from typing import Any, Mapping, NamedTuple, Type, TypeVar
import urllib

import idna
from schema import And, Optional, Or, Schema, SchemaError, Use
import toml

from xain_fl.logger import StructLogger, get_logger

logger: StructLogger = get_logger(__name__)


def error(key: str, description: str) -> str:
    """Return an error message for the given configuration item and
    description of the expected value type.

    Args:

        key (str): key of the configuration item
        description (str): description of the expected type of value
            for this configuration item
    """
    return f"Invalid `{key}`: value must be {description}"


def positive_integer(
    key: str, expected_value: str = "a strictly positive integer"
) -> Schema:
    """Return a validator for strictly positive integers for the given
    configuration item.

    Args:

        key (str): key of the configuration item
        expected_value (str): description of the expected type of
            value for this configuration item
    """
    return And(int, lambda value: value > 0, error=error(key, expected_value))


def non_negative_integer(
    key: str, expected_value: str = "a positive integer"
) -> Schema:
    """Return a non-negative integer validator for the given configuration
    item.

    Args:

        key: key of the configuration item
        expected_value: description of the expected type of
            value for this configuration item

    """
    return And(int, lambda value: value >= 0, error=error(key, expected_value))


def url(key: str, expected_value: str = "a valid URL") -> Schema:
    """Return a URL validator for the given configuration item.

    Args:

        key: key of the configuration item
        expected_value: description of the expected type of
            value for this configuration item

    """

    def is_valid_url(value):
        try:
            parsed = urllib.parse.urlparse(value)
        except (ValueError, urllib.error.URLError):
            return False
        # A URL is considered valid if it has at least a scheme and a
        # network location.
        return all([parsed.scheme, parsed.netloc])

    return And(str, is_valid_url, error=error(key, expected_value))


def is_valid_hostname(value: str) -> bool:
    """Return whether the given string is a valid hostname

    Args:

        value: string to check

    Returns:

        `True` if the given value is a valid hostname, `False`
        otherwise
    """
    try:
        idna.encode(value)
    except idna.IDNAError:
        return False
    return True


def is_valid_ip_address(value: str) -> bool:
    """Return whether the given string is a valid IP address

    Args:

        value: string to check

    Returns:

        `True` if the given value is a valid IP address, `False`
        otherwise
    """
    try:
        ipaddress.ip_address(value)
    except ipaddress.AddressValueError:
        return False
    return True


def hostname_or_ip_address(
    key: str, expected_value: str = "a valid domain name or IP address"
) -> Schema:
    """Return a hostname or IP address validator for the given
    configuration item.

    Args:

        key: key of the configuration item
        expected_value: description of the expected type of
            value for this configuration item

    """
    return And(
        str,
        Or(is_valid_hostname, is_valid_ip_address),
        error=error(key, expected_value),
    )


def log_level(key: str) -> Schema:
    """Return a validator for logging levels

    Args:

        key: key of the configuration item
    """
    log_levels = ["notset", "debug", "info", "warning", "error", "critical"]
    error_message = "one of: " + ", ".join(log_levels)
    log_level_validator = lambda value: value in log_levels
    return And(str, log_level_validator, error=error(key, error_message))


SERVER_SCHEMA = Schema(
    {
        Optional("host", default="localhost"): hostname_or_ip_address("server.host"),
        Optional("port", default=50051): non_negative_integer("server.port"),
        Optional("grpc_options", default=dict): Use(
            lambda opt: list(opt.items()),
            error=error("server.grpc_options", "valid gRPC options"),
        ),
    }
)

AI_SCHEMA = Schema(
    {
        "rounds": positive_integer("ai.rounds"),
        "epochs": non_negative_integer("ai.epochs"),
        Optional("min_participants", default=1): positive_integer(
            "ai.min_participants"
        ),
        Optional("fraction_participants", default=1.0): And(
            Or(int, float),
            lambda value: 0 < value <= 1.0,
            error=error("ai.fraction_participants", "a float between 0 and 1.0"),
        ),
    }
)

STORAGE_SCHEMA = Schema(
    {
        "endpoint": And(str, url, error=error("storage.endpoint", "a valid URL")),
        "bucket": Use(str, error=error("storage.endpoint", "an S3 bucket name")),
        "secret_access_key": Use(
            str, error=error("storage.secret_access_key", "a valid utf-8 string")
        ),
        "access_key_id": Use(
            str, error=error("storage.access_key_id", "a valid utf-8 string")
        ),
    }
)

LOGGING_SCHEMA = Schema(
    {Optional("level", default="info"): log_level("logging.level"),}
)


METRICS_SCHEMA = Schema(
    {
        Optional("enable", default=False): Use(
            bool, error=error("metrics.enable", "a boolean")
        ),
        Optional("host", default="localhost"): And(
            str,
            hostname_or_ip_address,
            error=error("metrics.host", "a valid hostname or ip address"),
        ),
        Optional("port", default=8086): non_negative_integer("metrics.port"),
        Optional("user", default="root"): Use(
            str, error=error("metrics.user", "a valid user")
        ),
        Optional("password", default="root"): Use(
            str, error=error("metrics.password", "a valid password")
        ),
        Optional("db_name", default="metrics"): Use(
            str, error=error("metrics.db_name", "a database name")
        ),
    }
)


CONFIG_SCHEMA = Schema(
    {
        Optional("server", default=SERVER_SCHEMA.validate({})): SERVER_SCHEMA,
        "ai": AI_SCHEMA,
        "storage": STORAGE_SCHEMA,
        Optional("logging", default=LOGGING_SCHEMA.validate({})): LOGGING_SCHEMA,
        Optional("metrics", default=METRICS_SCHEMA.validate({})): METRICS_SCHEMA,
    }
)


def create_class_from_schema(class_name: str, schema: Schema) -> Any:

    """Create a class named `class_name` from the given `schema`, where
    the attributes of the new class are the schema's keys.

    Args:

        class_name: name of the class to create
        schema: schema from which to create the class

    Returns:

        A new class where attributes are the given schema's keys
    """
    # pylint: disable=protected-access
    keys = schema._schema.keys()
    attributes = list(
        map(lambda key: key._schema if isinstance(key, Schema) else key, keys)
    )
    return namedtuple(class_name, attributes)


# pylint: disable=invalid-name
AiConfig = create_class_from_schema("AiConfig", AI_SCHEMA)
AiConfig.__doc__ = (
    "FL configuration: number of participant to each training round, etc."
)

ServerConfig = create_class_from_schema("ServerConfig", SERVER_SCHEMA)
ServerConfig.__doc__ = (
    "The server configuration: TLS, addresses for incoming connections, etc."
)

StorageConfig = create_class_from_schema("StorageConfig", STORAGE_SCHEMA)
StorageConfig.__doc__ = (
    "Storage related configuration: storage endpoints and credentials, etc."
)

LoggingConfig = create_class_from_schema("LoggingConfig", LOGGING_SCHEMA)
LoggingConfig.__doc__ = "Logging related configuration: log level, colors, etc."

MetricsConfig = create_class_from_schema("MetricsConfig", METRICS_SCHEMA)
MetricsConfig.__doc__ = (
    "Metrics related configuration: InfluxDB host, InfluxDB port, etc."
)

T = TypeVar("T", bound="Config")


class Config:
    """The coordinator configuration.

    Configuration is split in three sections: `Config.ai` for items
    directly related to the FL protocol, `Config.server` for the
    server infrastructure, and `Config.storage` for storage related
    items.

    The configuration is usually loaded from a dictionary the `Config`
    attributes map to the dictionary keys.

    Args:

        ai: the configuration corresponding to the `[ai]` section of
            the toml config file

        storage: the configuration corresponding to the `[storage]`
            section of the toml config fil

        server: the configuration corresponding to the `[server]`
            section of the toml config file

        logging: the configuration corresponding to the `[logging]`
            section of the toml config file

        metrics: the configuration corresponding to the `[metrics]`
            section of the toml config file

    :Example:

    Here is a valid configuration:

    .. code-block:: toml

       # This section correspond to the `Config.server` attribute
       [server]

       # Address to listen on for incoming gRPC connections
       host = "[::]"
       # Port to listen on for incoming gRPC connections
       port = 50051


       # This section corresponds to the `Config.ai` attribute
       [ai]

       # Number of global rounds the model is going to be trained for. This
       # must be a positive integer.
       rounds = 1

       # Number of local epochs per round
       epochs = 1

       # Minimum number of participants to be selected for a round.
       min_participants = 1

       # Fraction of total clients that participate in a training round. This
       # must be a float between 0 and 1.
       fraction_participants = 1.0

       # This section corresponds to the `Config.storage` attribute
       [storage]

       # URL to the storage service to use
       endpoint = "http://localhost:9000"

       # Name of the bucket for storing the aggregated models
       bucket = "aggregated_weights"

       # AWS secret access to use to authenticate to the storage service
       secret_access_key = "my-secret"

       # AWS access key ID to use to authenticate to the storage service
       access_key_id = "my-key-id"

    This config file can be loaded and used as follows:

    .. code-block:: python

       from xain_fl.config import Config

       config = Config.load("example_config.toml")

       assert config.server.host == "[::]"
       assert config.server.port == 50051

       assert config.ai.rounds == 1
       assert config.ai.epochs == 1
       assert config.ai.min_participants == 1
       assert config.ai.fraction_participants == 1.0

       assert config.storage.endpoint == "http://localhost:9000"
       assert config.storage.bucket == "aggregated_weights"
       assert config.storage.secret_access_key == "my-access-key"
       assert config.storage.access_key_id == "my-key"
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        ai: NamedTuple,
        storage: NamedTuple,
        server: NamedTuple,
        logging: NamedTuple,
        metrics: NamedTuple,
    ):
        self.ai = ai
        self.storage = storage
        self.server = server
        self.logging = logging
        self.metrics = metrics

    @classmethod
    def from_unchecked_dict(cls: Type[T], dictionary: Mapping[str, Any]) -> T:
        """Check if the given dictionary contains a valid configuration, and
        if so, create a `Config` instance from it.

        Args:

            dictionary: a dictionary containing the configuration
        """
        try:
            valid_config = CONFIG_SCHEMA.validate(dictionary)
        except SchemaError as err:
            raise InvalidConfig(err.code) from err
        return cls.from_valid_dict(valid_config)

    @classmethod
    def from_valid_dict(cls: Type[T], dictionary: Mapping[str, Any]) -> T:
        """Create a `Config` instance for the given dictionary, assuming it
        contains a valid configuration

        Args:

            dictionary: a dictionary containing the configuration

        """
        return cls(
            AiConfig(**dictionary["ai"]),
            StorageConfig(**dictionary["storage"]),
            ServerConfig(**dictionary["server"]),
            LoggingConfig(**dictionary["logging"]),
            MetricsConfig(**dictionary["metrics"]),
        )

    @classmethod
    def load(cls: Type[T], path: str) -> T:
        """Read the config file from the given path, check that it contains a
        valid configuration, and return the corresponding `Config`
        instance.

        Args:

            path: path to a configuration file
        """
        try:
            with open(path, "r") as f:
                raw_config = toml.load(f)
        except IsADirectoryError as err:
            raise InvalidConfig(f"{path} is a directory") from err
        except FileNotFoundError as err:
            raise InvalidConfig(f"{path} not found") from err
        except PermissionError as err:
            raise InvalidConfig(
                f"failed to read {path}: insufficient permissions"
            ) from err
        except toml.TomlDecodeError as err:
            raise InvalidConfig(f"failed to decode {path}: {err}") from err
        except OSError as err:
            raise InvalidConfig(str(err)) from err
        return cls.from_unchecked_dict(raw_config)


class InvalidConfig(ValueError):
    """
    Exception raised upon trying to load an invalid configuration
    """



"""Contains coordinator logic which runs the central server in federated
learning. It consists of the following modules:

    - aggregate
    - controller
"""

from xain_fl.fl.coordinator.aggregate import Aggregator
from xain_fl.fl.coordinator.controller import RandomController



"""XAIN FL Rounds"""

from typing import Dict, List, Tuple

from numpy import ndarray

from xain_fl.tools.exceptions import DuplicatedUpdateError


class Round:
    """Class to manage the state of a single round.
    This class contains the logic to handle all updates sent by the
    participants during a round and does some sanity checks like preventing the
    same participant to submit multiple updates during a single round.

    Args:
        participant_ids: The list of IDs of the participants selected
            to participate in this round.
    """

    def __init__(self, participant_ids: List[str]) -> None:
        self.participant_ids = participant_ids
        self.updates: Dict[str, Dict] = {}

    def add_updates(
        self,
        participant_id: str,
        model_weights: ndarray,
        aggregation_data: int,
        metrics: Dict[str, ndarray],
    ) -> None:
        """Valid a participant's update for the round.

        Args:
            participant_id: The id of the participant making the request.

            model_weights: The updated model weights.

            aggregation_data: Meta data for aggregation.

            metrics: A dictionary containing metrics with the name and
                the value as list of ndarrays.

        Raises:
            DuplicatedUpdateError: If the participant already submitted his update this round.
        """

        if participant_id in self.updates.keys():
            raise DuplicatedUpdateError(
                f"Participant {participant_id} already submitted the update for this round."
            )

        self.updates[participant_id] = {
            "model_weights": model_weights,
            "aggregation_data": aggregation_data,
            "metrics": metrics,
        }

    def is_finished(self) -> bool:
        """Check if all the required participants submitted their updates this round.
        If all participants submitted their updates the round is considered finished.

        Returns:
            `True` if all participants submitted their updates this
            round. `False` otherwise.
        """

        return len(self.updates) == len(self.participant_ids)

    def get_weight_updates(self) -> Tuple[List[ndarray], List[int]]:
        """Get a list of all participants weight updates.
        This list will usually be used by the aggregation function.

        Returns:
            The lists of model weights and aggregation meta data from all participants.
        """

        return (
            [v["model_weights"] for v in self.updates.values()],
            [v["aggregation_data"] for v in self.updates.values()],
        )



"""Provides an abstract base class Controller and the RandomController
currently used by the Coordinator."""

from abc import ABC, abstractmethod
from typing import List

import numpy as np


# TODO: raise exceptions for invalid attribute values: https://xainag.atlassian.net/browse/XP-387
class Controller(ABC):
    """Abstract base class which provides an interface to the coordinator that
    enables different selection strategies.

    Attributes:
        fraction_of_participants (:obj:`float`, optional): The fraction of total
            participant IDs to be selected. Defaults to 1.0, meaning that
            all participant IDs will be selected. It must be in the (0.0, 1.0] interval.
    """

    def __init__(self, fraction_of_participants: float = 1.0) -> None:
        """[summary]

        .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

        Args:
            fraction_of_participants (float): [description]. Defaults to 1.0.
        """

        self.fraction_of_participants: float = fraction_of_participants

    def get_num_ids_to_select(self, len_participant_ids: int) -> int:
        """Calculates how many participant IDs need to be selected.

        Args:
            len_participant_ids (:obj:`int`): The length of the list of IDs of all the
                available participants.

        Returns:
            :obj:`int`: Number of participant IDs to be selected
        """
        raw_num_ids_to_select = len_participant_ids * self.fraction_of_participants
        max_valid_value = max(1, np.ceil(raw_num_ids_to_select))
        minimum_valid_value = min(len_participant_ids, max_valid_value)
        return int(minimum_valid_value)

    @abstractmethod
    def select_ids(self, participant_ids: List[str]) -> List[str]:
        """Returns the selected indices of next round.

        Args:
            participant_ids (:obj:`list` of :obj:`str`): The list of IDs of all the
                available participants, a subset of which will be selected.

        Returns:
            :obj:`list` of :obj:`str`: List of selected participant IDs
        """
        raise NotImplementedError("not implemented")


class IdController(Controller):
    """[summary

    ... todo: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)
    """

    def select_ids(self, participant_ids: List[str]) -> List[str]:
        """Selects all given participants.

        Args:
            participant_ids (:obj:`list` of :obj:`str`): The list of IDs of all the
                available participants, a subset of which will be selected.

        Returns:
            :obj:`list` of :obj:`str`: List of selected participant IDs
        """

        return participant_ids


class RandomController(Controller):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        Controller ([type]): [description]
    """

    def select_ids(self, participant_ids: List[str]) -> List[str]:
        """Randomly samples self.num_ids_to_select from the population of participants_ids,
        without replacement.

        Args:
            participant_ids (:obj:`list` of :obj:`str`): The list of IDs of all the
                available participants, a subset of which will be selected.

        Returns:
            :obj:`list` of :obj:`str`: List of selected participant IDs
        """

        num_ids_to_select = self.get_num_ids_to_select(len(participant_ids))
        return np.random.choice(participant_ids, size=num_ids_to_select, replace=False)



"""This module is the entrypoint to start a new coordinator instance.

"""
import sys

from xain_fl.config import Config, InvalidConfig, get_cmd_parameters
from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.coordinator.metrics_store import (
    AbstractMetricsStore,
    MetricsStore,
    NullObjectMetricsStore,
)
from xain_fl.coordinator.store import S3Store
from xain_fl.logger import StructLogger, get_logger, initialize_logging, set_log_level
from xain_fl.serve import serve

logger: StructLogger = get_logger(__name__)


def main():
    """Start a coordinator instance
    """
    initialize_logging()

    args = get_cmd_parameters()
    try:
        config = Config.load(args.config)
    except InvalidConfig as err:
        logger.error("Invalid config", error=str(err))
        sys.exit(1)

    set_log_level(config.logging.level.upper())

    metrics_store: AbstractMetricsStore = NullObjectMetricsStore()
    if config.metrics.enable:
        metrics_store = MetricsStore(config.metrics)

    coordinator = Coordinator(
        store=S3Store(config.storage),
        num_rounds=config.ai.rounds,
        epochs=config.ai.epochs,
        minimum_participants_in_round=config.ai.min_participants,
        fraction_of_participants=config.ai.fraction_participants,
        metrics_store=metrics_store,
    )

    serve(coordinator=coordinator, server_config=config.server)


if __name__ == "__main__":
    main()



"""XAIN FL Coordinator"""

from typing import Dict, List

from google.protobuf.internal.python_message import GeneratedProtocolMessageType
import numpy as np
from numpy import ndarray
from xain_proto.fl.coordinator_pb2 import (
    EndTrainingRoundRequest,
    EndTrainingRoundResponse,
    HeartbeatRequest,
    HeartbeatResponse,
    RendezvousReply,
    RendezvousRequest,
    RendezvousResponse,
    StartTrainingRoundRequest,
    StartTrainingRoundResponse,
    State,
)
from xain_proto.np import ndarray_to_proto, proto_to_ndarray

from xain_fl.coordinator.metrics_store import (
    AbstractMetricsStore,
    MetricsStoreError,
    NullObjectMetricsStore,
)
from xain_fl.coordinator.participants import Participants
from xain_fl.coordinator.round import Round
from xain_fl.coordinator.store import AbstractStore, NullObjectStore
from xain_fl.fl.coordinator.aggregate import Aggregator, WeightedAverageAggregator
from xain_fl.fl.coordinator.controller import Controller, RandomController
from xain_fl.logger import StructLogger, get_logger
from xain_fl.tools.exceptions import InvalidRequestError, UnknownParticipantError

logger: StructLogger = get_logger(__name__)


# TODO: raise exceptions for invalid attribute values: https://xainag.atlassian.net/browse/XP-387
class Coordinator:  # pylint: disable=too-many-instance-attributes
    """Class implementing the main Coordinator logic. It is implemented as a
    state machine that reacts to received messages.

    The states of the Coordinator are:
        - ``STANDBY``: The coordinator is in standby mode, typically
          when waiting for participants to connect. In this mode the
          only messages that the coordinator can receive are
          :class:`~.coordinator_pb2.RendezvousRequest` and
          :class:`~.coordinator_pb2.HeartbeatRequest`.

        - ``ROUND``: A round is currently in progress. During a round
          the important messages the coordinator can receive are
          :class:`~.coordinator_pb2.StartTrainingRoundRequest` and
          :class:`~.coordinator_pb2.EndTrainingRoundRequest`.  Since
          participants may or may not be selected for rounds, they can be
          advertised accordingly with ROUND or STANDBY respectively.
          Round numbers start from 0.

        - ``FINISHED``: The training session has ended and
          participants should disconnect from the coordinator.

    States are exchanged during heartbeats so that both coordinators
    and participants can react to each others state change.

    The flow of the Coordinator:
        1. The coordinator is started and waits for enough participants to join. `STANDBY`.
        2. Once enough participants are connected the coordinator starts the rounds. `ROUND`.
        3. Repeat step 2. for the given number of rounds
        4. The training session is over and the coordinator is ready to shutdown. `FINISHED`.

    Note:
        :class:`~.coordinator_pb2.RendezvousRequest` is always allowed
        regardless of which state the coordinator is on.

    Args:
        num_rounds: The number of rounds of the training session.

        minimum_participants_in_round: The minimum number of
            participants that participate in a round.

        fraction_of_participants: The fraction of total connected
            participants to be selected in a single round. Defaults to
            1.0, meaning that all connected participants will be
            selected. It must be in the (0.0, 1.0] interval.

        weights: The weights of the global model.

        epochs: Number of training iterations local to Participant.

        epochs_base: Global number of epochs as of last round.

        aggregator: The type of aggregation to perform at the end of
            each round. Defaults to :class:`~.WeightedAverageAggregator`.

        controller: Controls how the Participants are selected at the
            start of each round. Defaults to :class:`~.RandomController`.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        store: AbstractStore = NullObjectStore(),
        metrics_store: AbstractMetricsStore = NullObjectMetricsStore(),
        num_rounds: int = 1,
        minimum_participants_in_round: int = 1,
        fraction_of_participants: float = 1.0,
        weights: ndarray = np.empty(shape=(0,)),
        epochs: int = 1,
        epoch_base: int = 0,
        aggregator: Aggregator = WeightedAverageAggregator(),
        controller: Controller = RandomController(),
    ) -> None:
        self.store: AbstractStore = store
        self.minimum_participants_in_round: int = minimum_participants_in_round
        self.fraction_of_participants: float = fraction_of_participants
        self.participants: Participants = Participants()
        self.num_rounds: int = num_rounds
        self.aggregator: Aggregator = aggregator
        self.controller: Controller = controller
        self.metrics_store = metrics_store
        self.minimum_connected_participants: int = self.get_minimum_connected_participants()

        # global model
        self.weights: ndarray = weights
        self.epochs: int = epochs
        self.epoch_base: int = epoch_base

        # round updates
        self.round: Round = Round(self.participants.ids())

        # state variables
        self.state: State = State.STANDBY
        self.current_round: int = 0

    def get_minimum_connected_participants(self) -> int:
        """Calculates how many participants are needed so that we can select
        a specific fraction of them.

        Returns:
            obj:`int`: Minimum number of participants needed to be connected to start a round.
        """

        return int(self.minimum_participants_in_round // self.fraction_of_participants)

    def on_message(
        self, message: GeneratedProtocolMessageType, participant_id: str
    ) -> GeneratedProtocolMessageType:
        """Coordinator method that implements the state machine.

        Args:
            message: A protobuf message.
            participant_id: The id of the participant making the request.

        Returns:
            The response sent back to the participant.

        Raises:
            UnknownParticipantError: If it receives a request from an
                unknown participant. Typically a participant that has not
                rendezvous with the :class:`~.Coordinator`.

            InvalidRequestError: If it receives a request that is not
                allowed in the current :class:`~.Coordinator` state.
        """

        logger.debug(
            "Received message from participant",
            message_type=type(message),
            participant_id=participant_id,
        )

        # Unless this is a RendezvousRequest the coordinator should not accept messages
        # from participants that have not been accepted
        if (
            not isinstance(message, RendezvousRequest)
            and participant_id not in self.participants.ids()
        ):
            raise UnknownParticipantError(
                f"Unknown participant {participant_id}. "
                "Please try to rendezvous with the coordinator before making a request."
            )

        if isinstance(message, RendezvousRequest):
            # Handle rendezvous
            return self._handle_rendezvous(message, participant_id)

        if isinstance(message, HeartbeatRequest):
            # Handle heartbeat
            return self._handle_heartbeat(message, participant_id)

        if isinstance(message, StartTrainingRoundRequest):
            # handle start training
            return self._handle_start_training_round(message, participant_id)

        if isinstance(message, EndTrainingRoundRequest):
            # handle end training
            return self._handle_end_training_round(message, participant_id)

        raise NotImplementedError

    def remove_participant(self, participant_id: str) -> None:
        """Remove a participant from the list of accepted participants.

        This method is to be called when it is detected that a
        participant has disconnected. After a participant is removed,
        if the number of remaining participants is less than the
        minimum number of participants that need to be connected, the
        :class:`~.Coordinator` will transition to STANDBY state.

        Args:
            participant_id: The id of the participant to remove.
        """

        self.participants.remove(participant_id)
        logger.info("Removing participant", participant_id=participant_id)

        if self.participants.len() < self.minimum_connected_participants:
            self.state = State.STANDBY

    def select_participant_ids_and_init_round(self) -> None:
        """Selects the participant ids and initiates a Round."""

        self.controller.fraction_of_participants = self.fraction_of_participants
        selected_ids = self.controller.select_ids(self.participants.ids())
        self.round = Round(selected_ids)

    def _handle_rendezvous(
        self, _message: RendezvousRequest, participant_id: str
    ) -> RendezvousResponse:
        """Handles a Rendezvous request.

        Args:
            _message: The request to handle. Currently not used.
            participant_id: The id of the participant making the request.

        Returns:
            The response to the participant.
        """

        if self.participants.len() < self.minimum_connected_participants:
            reply = RendezvousReply.ACCEPT
            self.participants.add(participant_id)
            logger.info(
                "Accepted participant",
                participant_id=participant_id,
                current_participants_count=self.participants.len(),
            )

            # Select participants and change the state to ROUND if the latest added participant
            # lets us meet the minimum number of connected participants
            if self.participants.len() == self.minimum_connected_participants:
                self.select_participant_ids_and_init_round()

                self.state = State.ROUND
        else:
            reply = RendezvousReply.LATER
            logger.info(
                "Reject participant",
                participant_id=participant_id,
                current_participants_count=self.participants.len(),
            )

        return RendezvousResponse(reply=reply)

    def _handle_heartbeat(
        self, _message: HeartbeatRequest, participant_id: str
    ) -> HeartbeatResponse:
        """Handles a Heartbeat request.

        Responds to the participant with:
            - ``FINISHED``: if coordinator is in state FINISHED,
            - ``ROUND``: if the participant is selected for the current round,
            - ``STANDBY``: if the participant is not selected for the current round.

        Args:
            _message: The request to handle. Currently not used.
            participant_id: The id of the participant making the request.

        Returns:
            The response to the participant.
        """

        self.participants.update_expires(participant_id)

        if self.state == State.FINISHED or participant_id in self.round.participant_ids:
            state = self.state
        else:
            state = State.STANDBY

        # send heartbeat response advertising the current state
        logger.debug(
            "Heartbeat response",
            participant_id=participant_id,
            message=state,
            round=self.current_round,
        )
        return HeartbeatResponse(state=state, round=self.current_round)

    def _handle_start_training_round(
        self, _message: StartTrainingRoundRequest, participant_id: str
    ) -> StartTrainingRoundResponse:
        """Handles a StartTrainingRound request.

        Args:
            _message: The request to handle. Currently not used.
            participant_id: The id of the participant making the request.

        Returns:
            :class:`~.coordinator_pb2.StartTrainingRoundResponse`: The response to the participant.
        """

        # The coordinator should only accept StartTrainingRound requests if it is
        # in the ROUND state and when the participant has been selected for the round.
        coordinator_not_in_a_round = self.state != State.ROUND
        participant_not_selected = participant_id not in self.round.participant_ids
        if coordinator_not_in_a_round or participant_not_selected:
            raise InvalidRequestError(
                f"Participant {participant_id} sent a "
                "StartTrainingRoundRequest outside of a round"
            )

        return StartTrainingRoundResponse(
            weights=ndarray_to_proto(self.weights),
            epochs=self.epochs,
            epoch_base=self.epoch_base,
        )

    def _handle_end_training_round(
        self, message: EndTrainingRoundRequest, participant_id: str
    ) -> EndTrainingRoundResponse:
        """Handles a EndTrainingRound request.

        Args:
            message: The request to handle.
            participant_id: The id of the participant making the request.

        Returns:
            The response to the participant.
        """

        # TODO: Ideally we want to know for which round the participant is
        # submitting the updates and raise an exception if it is the wrong
        # round.

        # record the request data
        model_weights: ndarray = proto_to_ndarray(message.weights)
        number_samples: int = message.number_samples
        metrics: Dict[str, ndarray] = {
            k: proto_to_ndarray(v) for k, v in message.metrics.items()
        }
        self.round.add_updates(
            participant_id=participant_id,
            model_weights=model_weights,
            aggregation_data=number_samples,
            metrics=metrics,
        )

        try:
            self.metrics_store.write_metrics(participant_id, metrics)
        except MetricsStoreError as err:
            logger.warn(
                "Can not write metrics", participant_id=participant_id, error=repr(err)
            )

        # The round is over. Run the aggregation
        if self.round.is_finished():
            logger.info(
                "Running aggregation for round", current_round=self.current_round
            )

            multiple_model_weights: List[ndarray]
            aggregation_data: List[int]
            multiple_model_weights, aggregation_data = self.round.get_weight_updates()
            self.weights = self.aggregator.aggregate(
                multiple_model_weights=multiple_model_weights,
                aggregation_data=aggregation_data,
            )
            self.store.write_weights(round=self.current_round, weights=self.weights)

            # update the round or finish the training session
            if self.current_round >= self.num_rounds - 1:
                logger.info("Last round over", round=self.current_round)
                self.state = State.FINISHED
            else:
                self.current_round += 1
                # reinitialize the round
                self.select_participant_ids_and_init_round()

        return EndTrainingRoundResponse()



"""This module provides helpers for parsing the CLI arguments.
"""
import argparse


def get_cmd_parameters() -> argparse.Namespace:
    """Parse the command arguments

    Returns:
        ~argparse.Namespace: the parsed command arguments
    """
    parser = argparse.ArgumentParser(description="Coordinator CLI")
    parser.add_argument(
        "--config",
        dest="config",
        default="xain-fl.toml",
        help="Path to the coordinator configuration file",
    )
    return parser.parse_args()



"""XAIN FL tests for gRPC coordinator"""

from concurrent import futures
import threading
import time
from unittest import mock

import grpc
import numpy as np
import pytest
from xain_proto.fl.coordinator_pb2 import (
    EndTrainingRoundRequest,
    EndTrainingRoundResponse,
    HeartbeatRequest,
    HeartbeatResponse,
    RendezvousReply,
    RendezvousRequest,
    StartTrainingRoundRequest,
    StartTrainingRoundResponse,
    State,
)
from xain_proto.fl.coordinator_pb2_grpc import add_CoordinatorServicer_to_server
from xain_proto.np import ndarray_to_proto
from xain_sdk.participant_state_machine import (
    StateRecord,
    end_training_round,
    message_loop,
    rendezvous,
    start_participant,
    start_training_round,
)

from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.coordinator.coordinator_grpc import CoordinatorGrpc
from xain_fl.coordinator.heartbeat import monitor_heartbeats
from xain_fl.coordinator.participants import ParticipantContext, Participants

from .store import MockS3Store


@pytest.mark.integration
def test_participant_rendezvous_accept(  # pylint: disable=unused-argument
    participant_stub, coordinator_service
):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
        coordinator_service ([type]): [description]
    """

    response = participant_stub.Rendezvous(RendezvousRequest())

    assert response.reply == RendezvousReply.ACCEPT


@pytest.mark.integration
def test_participant_rendezvous_later(participant_stub):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
    """

    # populate participants
    coordinator = Coordinator(
        minimum_participants_in_round=10, fraction_of_participants=1.0
    )
    required_participants = 10
    for i in range(required_participants):
        coordinator.participants.add(str(i))

    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    add_CoordinatorServicer_to_server(CoordinatorGrpc(coordinator), server)
    server.add_insecure_port("localhost:50051")
    server.start()

    # try to rendezvous the 11th participant
    response = participant_stub.Rendezvous(RendezvousRequest())
    server.stop(0)

    assert response.reply == RendezvousReply.LATER


@pytest.mark.integration
def test_heartbeat(participant_stub, coordinator_service):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
        coordinator_service ([type]): [description]
    """

    # first we need to rendezvous so that the participant is added to the list of participants
    _ = participant_stub.Rendezvous(RendezvousRequest())
    response = participant_stub.Heartbeat(HeartbeatRequest())

    # the Coordinator is initialised in conftest.py::coordinator_service with 10 participants
    # needed per round. so here we expect the HeartbeatResponse to have State.STANDBY
    # because we connected only one participant
    assert response == HeartbeatResponse()
    assert coordinator_service.coordinator.state == State.STANDBY


@pytest.mark.integration
def test_heartbeat_denied(
    participant_stub, coordinator_service
):  # pylint: disable=unused-argument
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
        coordinator_service ([type]): [description]
    """

    # heartbeat requests are only allowed if the participant has already
    # rendezvous with the coordinator
    with pytest.raises(grpc.RpcError):
        response = participant_stub.Heartbeat(HeartbeatRequest())
        assert response.status_code == grpc.StatusCode.PERMISSION_DENIED


@mock.patch(
    "xain_fl.coordinator.heartbeat.threading.Event.is_set", side_effect=[False, True]
)
@mock.patch("xain_fl.coordinator.heartbeat.threading.Event.wait", return_value=None)
@mock.patch("xain_fl.coordinator.heartbeat.Coordinator.remove_participant")
def test_monitor_heartbeats(
    mock_participants_remove, _mock_event_wait, _mock_event_is_set
):
    """Test that when there is a participant with an expired heartbeat,
    ``Coordinator.remove_participant`` is called exactly once.

    Args:
        mock_participants_remove: mock of ``Coordinator.remove_participant()``
        _mock_event_wait: mock of ``threading.Event.wait`` that does not block
        _mock_event_is_set: mock of ``threading.Event.is_set``

    """

    participants = Participants()
    participants.add("participant_1")
    participants.participants["participant_1"].heartbeat_expires = 0

    coordinator = Coordinator()
    coordinator.participants = participants

    terminate_event = threading.Event()
    monitor_heartbeats(coordinator, terminate_event)

    mock_participants_remove.assert_called_once_with("participant_1")


@mock.patch(
    "xain_fl.coordinator.heartbeat.threading.Event.is_set", side_effect=[False, True]
)
@mock.patch("xain_fl.coordinator.heartbeat.threading.Event.wait", return_value=None)
def test_monitor_heartbeats_remove_participant(_mock_event_wait, _mock_event_is_set):
    """Test that when the coordinator has exactly one participant with an
    expired heartbeat, it is removed correctly.

    Args:

        _mock_event_wait: mock of ``threading.Event.wait`` that does not block
        _mock_event_is_set: mock of ``threading.Event.is_set``

    """

    participants = Participants()
    participants.add("participant_1")
    participants.participants["participant_1"].heartbeat_expires = 0

    coordinator = Coordinator()
    coordinator.participants = participants

    terminate_event = threading.Event()
    monitor_heartbeats(coordinator, terminate_event)

    assert participants.len() == 0


@pytest.mark.slow
def test_many_heartbeats_expire_in_short_interval():
    """Make sure that heartbeat_monitor() works correctly under heavy
    load. This test was added to reproduce
    https://xainag.atlassian.net/browse/PB-104

    """
    participants = {}
    for i in range(0, 100):
        participant = ParticipantContext(str(i))
        participant.heartbeat_expires = time.time() + 0.1 + i / 1000
        participants[str(i)] = participant
    coordinator = Coordinator()
    coordinator.participants.participants = participants

    terminate_event = threading.Event()

    def terminate_heartbeats_monitor():
        time.sleep(0.2)
        terminate_event.set()

    supervisor = threading.Thread(target=terminate_heartbeats_monitor, daemon=True)
    supervisor.start()

    monitor_heartbeats(coordinator, terminate_event)

    supervisor.join(timeout=0.1)
    assert not participants


@mock.patch(
    "xain_sdk.participant_state_machine.threading.Event.is_set",
    side_effect=[False, False, True],
)
@mock.patch("xain_sdk.participant_state_machine.time.sleep", return_value=None)
@mock.patch("xain_sdk.participant_state_machine.HeartbeatRequest")
def test_message_loop(mock_heartbeat_request, _mock_sleep, _mock_event):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        mock_heartbeat_request ([type]): [description]
        _mock_sleep ([type]): [description]
        _mock_event ([type]): [description]
    """

    channel = mock.MagicMock()
    terminate_event = threading.Event()
    state_record = StateRecord()

    message_loop(channel, state_record, terminate_event)

    # check that the heartbeat is sent exactly twice
    mock_heartbeat_request.assert_has_calls([mock.call(), mock.call()])


@pytest.mark.integration
def test_start_training_round(coordinator_service):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        coordinator_service ([type]): [description]
    """

    test_weights = np.arange(10)

    # set coordinator global model and hyper-params so that it needs only 1 participant
    coord = coordinator_service.coordinator
    coord.minimum_participants_in_round = 1
    coord.fraction_of_participants = 1.0
    coord.epochs = 5
    coord.epoch_base = 2
    coord.weights = test_weights
    coord.minimum_connected_participants = coord.get_minimum_connected_participants()

    # simulate a participant communicating with coordinator via channel
    with grpc.insecure_channel("localhost:50051") as channel:
        # we need to rendezvous before we can send any other requests
        rendezvous(channel)
        # call StartTrainingRound service method on coordinator
        weights, epochs, epoch_base = start_training_round(channel)

    # check global model received
    assert epochs == 5
    assert epoch_base == 2
    np.testing.assert_equal(weights, test_weights)


@pytest.mark.integration
def test_start_training_round_denied(  # pylint: disable=unused-argument
    participant_stub, coordinator_service
):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
        coordinator_service ([type]): [description]
    """

    # start training requests are only allowed if the participant has already
    # rendezvous with the coordinator
    with pytest.raises(grpc.RpcError):
        response = participant_stub.StartTrainingRound(StartTrainingRoundRequest())
        assert response.status_code == grpc.StatusCode.PERMISSION_DENIED


@pytest.mark.integration
def test_start_training_round_failed_precondition(  # pylint: disable=unused-argument
    participant_stub, coordinator_service
):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
        coordinator_service ([type]): [description]
    """

    # start training requests are only allowed if the coordinator is in ROUND state.
    # Since we need 10 participants to be connected (see conftest.py::coordinator_service)
    # the StartTrainingRoundRequest is expected to fail
    participant_stub.Rendezvous(RendezvousRequest())
    with pytest.raises(grpc.RpcError):
        response = participant_stub.StartTrainingRound(StartTrainingRoundRequest())
        assert response.status_code == grpc.StatusCode.FAILED_PRECONDITION


@pytest.mark.integration
def test_end_training_round(coordinator_service):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        coordinator_service ([type]): [description]
    """

    assert coordinator_service.coordinator.round.updates == {}

    # simulate trained local model data
    test_weights = np.arange(20)
    number_samples = 2
    test_metrics = {"metric": np.arange(10, 20)}

    with grpc.insecure_channel("localhost:50051") as channel:
        # we first need to rendezvous before we can send any other request
        rendezvous(channel)
        # call EndTrainingRound service method on coordinator
        end_training_round(channel, test_weights, number_samples, test_metrics)
    # check local model received...

    assert len(coordinator_service.coordinator.round.updates) == 1

    round_ = coordinator_service.coordinator.round

    # first the weights update
    _, round_update = round_.updates.popitem()
    np.testing.assert_equal(round_update["model_weights"], test_weights)
    assert round_update["aggregation_data"] == number_samples

    round_update_metrics = round_update["metrics"]
    assert round_update_metrics.keys() == test_metrics.keys()
    for key, values in test_metrics.items():
        np.testing.assert_equal(round_update_metrics[key], values)


@pytest.mark.integration
def test_end_training_round_duplicated_updates(  # pylint: disable=unused-argument
    coordinator_service, participant_stub
):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        coordinator_service ([type]): [description]
        participant_stub ([type]): [description]
    """

    # participant can only send updates once in a single round
    participant_stub.Rendezvous(RendezvousRequest())

    participant_stub.EndTrainingRound(EndTrainingRoundRequest())

    with pytest.raises(grpc.RpcError):
        response = participant_stub.EndTrainingRound(EndTrainingRoundRequest())
        assert response.status_code == grpc.StatusCode.ALREADY_EXISTS


@pytest.mark.integration
def test_end_training_round_denied(  # pylint: disable=unused-argument
    participant_stub, coordinator_service
):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        participant_stub ([type]): [description]
        coordinator_service ([type]): [description]
    """

    # heartbeat requests are only allowed if the participant has already
    # rendezvous with the coordinator
    with pytest.raises(grpc.RpcError):
        response = participant_stub.EndTrainingRound(EndTrainingRoundRequest())
        assert response.status_code == grpc.StatusCode.PERMISSION_DENIED


@pytest.mark.integration
def test_full_training_round(participant_stubs, coordinator_service):
    """Run a complete training round with multiple participants.
    """
    # Use a MockS3Store so that we can also test the storage logic
    coordinator_service.coordinator.store = MockS3Store()

    # Initialize the coordinator with dummy weights, otherwise, the
    # aggregated weights at the end of the round are an empty array.
    dummy_weights = np.array([1, 2, 3, 4])
    coordinator_service.coordinator.weights = dummy_weights
    weights_proto = ndarray_to_proto(dummy_weights)

    # Create 10 partipants
    participants = [next(participant_stubs) for _ in range(0, 10)]

    # 9 participants out of 10 connect to the coordinator, so it stays
    # in STANDBY and accepts the connections.
    for participant in participants[:-1]:
        response = participant.Rendezvous(RendezvousRequest())
        assert response.reply == RendezvousReply.ACCEPT

        response = participant.Heartbeat(HeartbeatRequest())
        assert response == HeartbeatResponse(state=State.STANDBY, round=0)

    assert coordinator_service.coordinator.state == State.STANDBY
    assert coordinator_service.coordinator.current_round == 0

    # The 10th participant connects, so the coordinator switches to ROUND
    last_participant = participants[-1]
    response = last_participant.Rendezvous(RendezvousRequest())
    assert response.reply == RendezvousReply.ACCEPT

    assert coordinator_service.coordinator.state == State.ROUND
    assert coordinator_service.coordinator.current_round == 0

    response = last_participant.Heartbeat(HeartbeatRequest())
    assert response == HeartbeatResponse(state=State.ROUND, round=0)

    # The initial 9 participants send another heartbeat request.
    for participant in participants[:-1]:
        response = participant.Heartbeat(HeartbeatRequest(state=State.STANDBY, round=0))
    assert response == HeartbeatResponse(state=State.ROUND, round=0)

    # The participants start training
    for participant in participants:
        response = participant.StartTrainingRound(StartTrainingRoundRequest())
        assert response == StartTrainingRoundResponse(
            weights=weights_proto,
            epochs=coordinator_service.coordinator.epochs,
            epoch_base=coordinator_service.coordinator.epoch_base,
        )

    # The first 9 participants end training
    for participant in participants[:-1]:
        response = participant.EndTrainingRound(
            EndTrainingRoundRequest(weights=weights_proto, number_samples=1)
        )
        assert response == EndTrainingRoundResponse()

    assert not coordinator_service.coordinator.round.is_finished()
    coordinator_service.coordinator.store.assert_didnt_write(1)

    # The last participant finishes training
    response = last_participant.EndTrainingRound(
        EndTrainingRoundRequest(weights=weights_proto, number_samples=1)
    )
    assert response == EndTrainingRoundResponse()
    # Make sure we wrote the results for the given round
    coordinator_service.coordinator.store.assert_wrote(
        0, coordinator_service.coordinator.weights
    )


@pytest.mark.integration
@pytest.mark.slow
def test_start_participant(mock_coordinator_service):
    """[summary]

    .. todo:: Advance docstrings (https://xainag.atlassian.net/browse/XP-425)

    Args:
        mock_coordinator_service ([type]): [description]
    """

    init_weight = np.arange(10)
    mock_coordinator_service.coordinator.weights = init_weight

    # mock a local participant with a constant train_round function
    with mock.patch("xain_sdk.participant_state_machine.Participant") as mock_obj:
        mock_local_part = mock_obj.return_value
        mock_local_part.train_round.return_value = init_weight, 1, {}

        start_participant(mock_local_part, "localhost:50051")

        coord = mock_coordinator_service.coordinator
        assert coord.state == State.FINISHED

        # coordinator set to 2 rounds for good measure, but the resulting
        # aggregated weights are the same as a single round
        assert coord.current_round == 1

        # expect weight aggregated by summation - see mock_coordinator_service
        np.testing.assert_equal(coord.weights, init_weight)



"""XAIN FL exceptions"""


class DuplicatedUpdateError(Exception):
    """Exception raised when the same participant tries to submit multiple
    updates to the :class:`xain_fl.coordinator.coordinator.Coordinator` in the same :class:`~.Round`
    """


class UnknownParticipantError(Exception):
    """Exception raised when a participant that is unknown to the
    :class:`xain_fl.coordinator.coordinator.Coordinator` makes a request.

    Typically this means that a participant tries to make a request before it
    has successfully rendezvous with the :class:`xain_fl.coordinator.coordinator.Coordinator`.
    """


class InvalidRequestError(Exception):
    """Exception raised when the Coordinator receives and invalid request from a participant.

    This can happen if the participant sends a request that is not allowed in a
    give Coordinator state. For instance the Coordinator will only accept
    StartTrainingRound requests during a ROUND.
    """



"""This package provides the logic for reading and validating the
various configuration options exposed by the CLI and the toml config
file."""

from xain_fl.config.cli import get_cmd_parameters
from xain_fl.config.schema import (
    AiConfig,
    Config,
    InvalidConfig,
    LoggingConfig,
    MetricsConfig,
    ServerConfig,
    StorageConfig,
)

__all__ = [
    "get_cmd_parameters",
    "Config",
    "AiConfig",
    "LoggingConfig",
    "StorageConfig",
    "ServerConfig",
    "MetricsConfig",
    "InvalidConfig",
]



"""XAIN FL Hearbeats"""

import threading
import time
from typing import List

from xain_fl.coordinator.coordinator import Coordinator
from xain_fl.logger import StructLogger, get_logger

logger: StructLogger = get_logger(__name__)


def monitor_heartbeats(
    coordinator: Coordinator, terminate_event: threading.Event
) -> None:
    """Monitors the heartbeat of participants.

    If a heartbeat expires the participant is removed from the :class:`~.Participants`.

    Note:

        This is meant to be run inside a thread and expects an
        :class:`~threading.Event`, to know when it should terminate.

    Args:

        coordinator: The coordinator to monitor for heartbeats.

        terminate_event: A threading event to signal that this method
            should terminate.
    """

    logger.info("Heartbeat monitor starting...")

    while not terminate_event.is_set():
        participants_to_remove: List[str] = []

        now = time.time()
        for participant in coordinator.participants.participants.values():
            if participant.heartbeat_expires < now:
                participants_to_remove.append(participant.participant_id)

        for participant_id in participants_to_remove:
            coordinator.remove_participant(participant_id)

        next_expiration: float = coordinator.participants.next_expiration() - now

        logger.debug("Monitoring heartbeats", next_expiration=next_expiration)

        # It is better not to use time.sleep() here because if there
        # is no participant,
        # coordinator.participants.next_expiration() can be quite big
        # and we end up sleeping for a long time, even if
        # terminate_event gets set in the meantime.
        terminate_event.wait(timeout=next_expiration)



import os.path
import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 6):
    sys.exit("Please use Python version 3.6 or higher.")

project_dir = os.path.dirname(os.path.abspath(__file__))
version_file_path = os.path.join(project_dir, "xain_fl/__version__.py")
readme_file_path = os.path.join(project_dir, "README.md")

# get version
version = {}
with open(version_file_path) as fp:
    exec(fp.read(), version)


# get readme
with open(readme_file_path, "r") as fp:
    readme = fp.read()


# License comments according to `pip-licenses`

install_requires = [
    "numpy==1.15",  # BSD
    "grpcio==1.23",  # Apache License 2.0
    "structlog==19.2.0",  # Apache License 2.0
    "xain-proto==0.4.0",  # Apache License 2.0
    "boto3==1.10.48",  # Apache License 2.0
    "toml==0.10.0",  # MIT
    "schema~=0.7",  # MIT
    "idna==2.8",  # BSD
    "influxdb==5.2.3",  # MIT
]

dev_require = [
    "black==19.10b0",  # MIT
    "mypy==0.760",  # MIT License
    "pylint==2.3.1",  # GPL
    "astroid==2.2.5",  # LGPL
    "isort==4.3.21",  # MIT
    "pip-licenses==1.15.2",  # MIT License
    "twine==2.0.0",  # Apache License 2.0
    "wheel==0.33.6",  # MIT
    "typing-extensions==3.7.4.1",  # PSF
]

tests_require = [
    "pytest==5.3.2",  # MIT license
    "pytest-cov==2.8.1",  # MIT
    "pytest-watch==4.2.0",  # MIT
    "xain-sdk==0.4.0",  # Apache License 2.0
]

docs_require = [
    "Sphinx==2.2.1",
    "m2r==0.2.1",
    "sphinxcontrib-mermaid==0.3.1",
    "sphinx-autodoc-typehints==1.10.3",  # MIT
    # docutils is a transitive dependency for docs_requires and
    # install_requires, which are handled sequentially by pip. If the
    # docs_requires dependencies are handled first, `docutils~=0.16`
    # gets installed, whereas install_requires depends on
    # `docutils<0.16`. To prevent this from happening, we just pin the
    # docutils version here.
    "docutils < 0.16",
]

setup(
    name="xain_fl",
    version=version["__version__"],
    description="XAIN is an open source framework for federated learning.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/xainag/xain-fl",
    author=["XAIN AG"],
    author_email="services@xain.io",
    license="Apache License Version 2.0",
    zip_safe=False,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        "test": tests_require,
        "docs": docs_require,
        "dev": dev_require + tests_require + docs_require,
    },
    entry_points={"console_scripts": ["coordinator=xain_fl.__main__:main"]},
)



"""This module provides classes for weights storage. It currently only
works with services that provide the AWS S3 APIs.

"""
import abc
from io import BytesIO
import pickle

import boto3
from numpy import ndarray

from xain_fl.config import StorageConfig


class AbstractStore(abc.ABC):
    """An abstract class that defines the API a store must implement."""

    @abc.abstractmethod
    def write_weights(self, round: int, weights: ndarray) -> None:
        """Store the given `weights`, corresponding to the given `round`.

        Args:
            round: A round number the weights correspond to.
            weights: The weights to store.
        """

    @abc.abstractmethod
    def read_weights(self, participant_id: str, round: int) -> ndarray:
        """Read the weights computed by the given participant for the given round.

        Args:
            participant_id: ID of the participant's weights.
            round: A round number the weights correspond to.
        """


class NullObjectStore(AbstractStore):
    """A store that does nothing."""

    def write_weights(self, round: int, weights: ndarray) -> None:
        """A dummy method that has no effect.

        Args:
            round: A round number the weights correspond to. Not used.
            weights: The weights to store. Not used.
        """

    def read_weights(self, participant_id: str, round: int) -> ndarray:
        """A dummy method that has no effect.

        Args:
            participant_id: ID of the participant's weights. Not used.
            round: A round number the weights correspond to. Not used.
        """


class S3Store(AbstractStore):
    """A store for services that offer the AWS S3 API.

    Args:
        config: the storage configuration (endpoint URL, credentials, etc.)
    """

    def __init__(self, config: StorageConfig):
        self.config = config
        self.s3 = boto3.resource(  # pylint: disable=invalid-name
            "s3",
            endpoint_url=self.config.endpoint,
            aws_access_key_id=self.config.access_key_id,
            aws_secret_access_key=self.config.secret_access_key,
            # FIXME: not sure what this should be for now
            region_name="dummy",
        )

    def write_weights(self, round: int, weights: ndarray):
        """Store the given `weights`, corresponding to the given `round`.

        Args:
            round: A round number the weights correspond to.
            weights: The weights to store.
        """

        bucket = self.s3.Bucket(self.config.bucket)
        bucket.put_object(Body=pickle.dumps(weights), Key=str(round))

    def read_weights(self, participant_id: str, round: int) -> ndarray:
        """Download the weights computed by the given participant for the given
        round, from an S3 bucket.

        Args:

            participant_id: ID of the participant's weights. Not used.
            round: A round number the weights correspond to.

        Return:
            The weights read from the S3 bucket.
        """

        bucket = self.s3.Bucket(self.config.bucket)
        data = BytesIO()
        bucket.download_fileobj(str(round), data)
        # FIXME: not sure whether getvalue() copies the data. If so we
        # should probably prefer getbuffer() instead.
        return pickle.loads(data.getvalue())



"""XAIN FL Metric Store"""

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Dict, List

from influxdb import InfluxDBClient
from numpy import array2string, ndarray

from xain_fl.config import MetricsConfig


class AbstractMetricsStore(ABC):  # pylint: disable=too-few-public-methods
    """An abstract metric store."""

    @abstractmethod
    def write_metrics(self, participant_id: str, metrics: Dict[str, ndarray]) -> None:
        """Write the participant metrics on behalf of the participant with the given participant_id
        into a metric store.

        Args:

            participant_id: The ID of the participant.
            metrics: The metrics of the participant with the given participant_id.

        Returns:

            True, on success, otherwise False.
        """


class NullObjectMetricsStore(
    AbstractMetricsStore
):  # pylint: disable=too-few-public-methods
    """A metric store that does nothing."""

    def write_metrics(self, participant_id: str, metrics: Dict[str, ndarray]) -> None:
        """A method that has no effect.

        Args:

            participant_id: The ID of the participant.
            metrics: The metrics of the participant with the given participant_id.

        Returns:

            Always True.
        """
        return True


class MetricsStore(AbstractMetricsStore):  # pylint: disable=too-few-public-methods
    """A metric store that uses InfluxDB to store the metrics."""

    def __init__(self, config: MetricsConfig):
        self.config = config
        self.influx_client = InfluxDBClient(
            self.config.host,
            self.config.port,
            self.config.user,
            self.config.password,
            self.config.db_name,
        )

    def write_metrics(self, participant_id: str, metrics: Dict[str, ndarray]) -> None:
        """Write the participant metrics on behalf of the participant with the given participant_id
        into InfluxDB.

        Args:

            participant_id: The ID of the participant.
            metrics: The metrics of the participant with the given participant_id.

        Returns:

            True, on success, otherwise False.

        Raises:

            MetricsStoreError: If the writing of the metrics to InfluxDB failed.
        """

        # FIXME: We will change the data format of the metrics message in a separate ticket
        # (PB-390). The goal is, that coordinator doesn't need to transform the metrics anymore.
        influx_data_points = transform_metrics_to_influx_data_points(
            participant_id, metrics
        )

        try:
            self.influx_client.write_points(influx_data_points)
        except Exception as err:  # pylint: disable=broad-except
            raise MetricsStoreError("Can not write metrics.") from err


# Check if ths function can be removed after PB-390 is done.
def transform_metrics_to_influx_data_points(
    participant_id: str, metrics: Dict[str, ndarray]
) -> List[dict]:
    """Transform the metrics of a participant into InfluxDB data points.

    Args:
        participant_id: The ID of the participant.
        metrics: The metrics of the participant with the given participant_id.

    Returns:
        The metrics of the participant as InfluxDB data points.
    """

    # Currently the metrics message does not contain any timestamps.
    # We set a timestamp for each epoch data point with an interval of 1 sec for the case,
    # that the participant send metrics for more than 1 epoch per round.
    # If we just use the same timestamp, all data points will group on a horizontal line.
    first_epoch_time = datetime.now()
    data_points: List = []

    for metric_name, epoch_data_points in metrics.items():
        next_epoch_time = first_epoch_time

        for epoch_data_point in epoch_data_points:
            data_point = {
                "measurement": f"participant.ai.{metric_name}",
                "tags": {"id": participant_id},
                "time": next_epoch_time,
                "fields": {
                    metric_name: array2string(
                        epoch_data_point,
                        precision=8,
                        suppress_small=True,
                        floatmode="fixed",
                    )
                },
            }

            data_points.append(data_point)
            next_epoch_time += timedelta(seconds=1)

    return data_points


class MetricsStoreError(Exception):
    """
    Raised when the writing of the metrics failed.
    """



"""XAIN FL tests for coordinator aggregation"""

# TODO: (XP-351) decide how to continue with that
# import numpy as np
#
# from xain_fl.fl.coordinator.aggregate import federated_averaging
#
#
# def test_federated_averaging():  # pylint: disable=too-many-locals
#     """Test for `federated_averaging()`."""
#
#     # Prepare:
#     # - Three weight updates (u0, u1, u2)
#     # - One layer in the model
#     # - Two weight tensors in the layer (e.g. weights + bias)
#
#     u0_ = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 1.0, 1.0])
#     u1_ = np.array([2.0, 3.0, 1.0, 4.0, 5.0, 6.0, 1.0, 1.0])
#     u2_ = np.array([3.0, 1.0, 2.0, 4.0, 5.0, 6.0, 1.0, 1.0])
#     multiple_model_weights = [u0_, u1_, u2_]
#
#     model_weights_expected = np.array([2.0, 2.0, 2.0, 4.0, 5.0, 6.0, 1.0, 1.0])
#
#     weighting = np.ones((len(multiple_model_weights)))
#
#     # Execute
#     model_weights_actual = federated_averaging(multiple_model_weights, weighting)
#
#     # Assert
#     assert len(model_weights_actual) == len(model_weights_expected)
#     np.testing.assert_array_equal(model_weights_actual, model_weights_expected)
